@@ -14,7 +14,7 @@ from .margins import Margin
 from .screen import Point, WritePosition, _CHAR_CACHE
 from .utils import token_list_to_text, explode_tokens
 from prompt_toolkit.cache import SimpleCache
-from prompt_toolkit.filters import to_cli_filter
+from prompt_toolkit.filters import to_cli_filter, ViInsertMode, EmacsInsertMode
 from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
 from prompt_toolkit.reactive import Integer
 from prompt_toolkit.token import Token
@@ -834,6 +834,9 @@ class ColorColumn(object):
         self.token = token
 
 
+_in_insert_mode = ViInsertMode() | EmacsInsertMode()
+
+
 class Window(Container):
     """
     Container that holds a control.
@@ -1043,6 +1046,8 @@ class Window(Container):
 
     def _get_digraph_char(self, cli):
         " Return `False`, or the Digraph symbol to be used. "
+        if cli.quoted_insert:
+            return '^'
         if cli.vi_state.waiting_for_digraph:
             if cli.vi_state.digraph_symbol1:
                 return cli.vi_state.digraph_symbol1
@@ -1190,6 +1195,7 @@ class Window(Container):
         line_count = ui_content.line_count
         new_buffer = new_screen.data_buffer
         empty_char = _CHAR_CACHE['', Token]
+        ZeroWidthEscape = Token.ZeroWidthEscape
 
         # Map visible line number to (row, col) of input.
         # 'col' will always be zero if line wrapping is off.
@@ -1218,9 +1224,15 @@ class Window(Container):
                 x = -horizontal_scroll
 
                 visible_line_to_row_col[y] = (lineno, horizontal_scroll)
+                new_buffer_row = new_buffer[y + ypos]
 
                 for token, text in line:
-                    new_buffer_row = new_buffer[y + ypos]
+                    # Remember raw VT escape sequences. (E.g. FinalTerm's
+                    # escape sequences.)
+                    if token == ZeroWidthEscape:
+                        new_screen.zero_width_escapes[y + ypos][x + xpos] += text
+                        continue
+
                     for c in text:
                         char = _CHAR_CACHE[c, token]
                         char_width = char.width
@@ -1306,6 +1318,11 @@ class Window(Container):
                     cli, new_screen, screen_cursor_position, xpos, ypos, width,
                     write_position.height)
 
+        # Draw input characters from the input processor queue.
+        if has_focus and ui_content.cursor_position:
+            self._show_input_processor_key_buffer(cli, new_screen)
+
+        # Set menu position.
         if not new_screen.menu_position and ui_content.menu_position:
             new_screen.menu_position = cursor_pos_to_screen_pos(
                     ui_content.menu_position.y, ui_content.menu_position.x)
@@ -1326,6 +1343,27 @@ class Window(Container):
             new_screen.data_buffer[cpos.y][cpos.x] = \
                 _CHAR_CACHE[digraph_char, Token.Digraph]
 
+    def _show_input_processor_key_buffer(self, cli, new_screen):
+        """
+        When the user is typing a key binding that consists of several keys,
+        display the last pressed key if the user is in insert mode and the key
+        is meaningful to be displayed.
+        E.g. Some people want to bind 'jj' to escape in Vi insert mode. But the
+             first 'j' needs to be displayed in order to get some feedback.
+        """
+        key_buffer = cli.input_processor.key_buffer
+
+        if key_buffer and _in_insert_mode(cli) and not cli.is_done:
+            # The textual data for the given key. (Can be a VT100 escape
+            # sequence.)
+            data = key_buffer[-1].data
+
+            # Display only if this is a 1 cell width character.
+            if get_cwidth(data) == 1:
+                cpos = new_screen.cursor_position
+                new_screen.data_buffer[cpos.y][cpos.x] = \
+                    _CHAR_CACHE[data, Token.PartialKeyBinding]
+
     def _highlight_cursorlines(self, cli, new_screen, cpos, x, y, width, height):
         """
         Highlight cursor row/column.
